from pyramid.config import Configurator


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    with Configurator(settings=settings) as config:
        config.include('twitcher.models')
        config.include('twitcher.frontpage')
        config.include('twitcher.oauth2')
        config.include('twitcher.api')
        config.include('twitcher.owsproxy')
        config.scan()
    return config.make_wsgi_app()



import pytest
import unittest

from pyramid import testing
from pyramid.testing import DummyRequest

from twitcher.owsrequest import OWSRequest
from twitcher.owsexceptions import OWSInvalidParameterValue


class OWSRequestWmsTestCase(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()

    def tearDown(self):
        testing.tearDown()

    def test_get_getcaps_request(self):
        params = dict(request="GetCapabilities", service="WMS", version="1.1.1")
        request = DummyRequest(params=params)
        ows_req = OWSRequest(request)
        assert ows_req.request == 'getcapabilities'
        assert ows_req.service == 'wms'
        assert ows_req.version == '1.1.1'
        assert ows_req.public_access
        assert ows_req.service_allowed

    def test_get_getcaps_request_downcase(self):
        params = dict(REQUEST="getcapabilities", SERVICE="WMS", VERSION="1.3.0")
        request = DummyRequest(params=params)
        ows_req = OWSRequest(request)
        assert ows_req.request == 'getcapabilities'
        assert ows_req.service == 'wms'
        assert ows_req.version == '1.3.0'

    def test_get_invalid_request(self):
        params = dict(REQUEST="givememore", SERVICE="WMS", VERSION="1.3.0")
        request = DummyRequest(params=params)

        with pytest.raises(OWSInvalidParameterValue):
            OWSRequest(request)

    def test_get_getmetadata_request(self):
        params = dict(REQUEST="GetMetadata", SERVICE="WMS", VERSION="1.3.0")
        request = DummyRequest(params=params)
        ows_req = OWSRequest(request)
        assert ows_req.request == 'getmetadata'
        assert ows_req.service == 'wms'
        assert ows_req.version == '1.3.0'



"""
Testing the OWS Registry.
"""

from .common import BaseTest, dummy_request

from twitcher.store import ServiceStore
from twitcher.owsregistry import OWSRegistry


class OWSRegistryTest(BaseTest):

    def setUp(self):
        super(OWSRegistryTest, self).setUp()
        self.init_database()

        service_store = ServiceStore(
            dummy_request(dbsession=self.session))

        self.reg = OWSRegistry(servicestore=service_store)

    def test_register_service_and_unregister_it(self):
        service = {
            'url': 'http://localhost/wps',
            'name': 'test_emu',
            'type': 'wps',
            'auth': 'token',
            'public': False,
            'verify': True,
            'purl': 'http://myservice/wps'}
        # register
        resp = self.reg.register_service(**service)
        assert resp == service

        # get by name
        resp = self.reg.get_service_by_name(service['name'])
        assert resp == service

        # list
        resp = self.reg.list_services()
        assert resp == [service]

        # unregister
        resp = self.reg.unregister_service(service['name'])
        assert resp is True

        # clear
        resp = self.reg.clear_services()
        assert resp is True



"""dev

Revision ID: 1ab5b6358dd5
Revises: d39766e64a84
Create Date: 2019-06-05 16:25:36.793494

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1ab5b6358dd5'
down_revision = 'd39766e64a84'
branch_labels = None
depends_on = None

def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('client',
    sa.Column('client_id', sa.String(length=40), nullable=False),
    sa.Column('client_secret', sa.String(length=55), nullable=False),
    sa.PrimaryKeyConstraint('client_id', name=op.f('pk_client'))
    )
    op.create_index(op.f('ix_client_client_secret'), 'client', ['client_secret'], unique=True)
    op.create_table('token',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('client_id', sa.String(length=40), nullable=False),
    sa.Column('token_type', sa.String(length=40), nullable=True),
    sa.Column('access_token', sa.String(length=255), nullable=True),
    sa.Column('refresh_token', sa.String(length=255), nullable=True),
    sa.Column('expires', sa.DateTime(), nullable=True),
    sa.Column('scope', sa.Text(), nullable=True),
    sa.ForeignKeyConstraint(['client_id'], ['client.client_id'], name=op.f('fk_token_client_id_client'), ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_token'))
    )
    op.drop_index('token_index', table_name='access_tokens')
    op.drop_table('access_tokens')
    # ### end Alembic commands ###

def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('access_tokens',
    sa.Column('id', sa.INTEGER(), nullable=False),
    sa.Column('token', sa.TEXT(), nullable=True),
    sa.Column('expires_at', sa.INTEGER(), nullable=True),
    sa.PrimaryKeyConstraint('id', name='pk_access_tokens')
    )
    op.create_index('token_index', 'access_tokens', ['token'], unique=1)
    op.drop_table('token')
    op.drop_index(op.f('ix_client_client_secret'), table_name='client')
    op.drop_table('client')
    # ### end Alembic commands ###



"""
Twitcher interfaces to allow alternative implementations in adapters.
"""
from typing import Any, Dict, List

from pyramid.request import Request

from twitcher.models import Service


class OWSSecurityInterface(object):
    def verify_request(self, request) -> bool:
        """Verify that the service request is allowed."""
        raise NotImplementedError


class ServiceStoreInterface(object):
    def __init__(self, request: Request) -> None:
        self.request = request

    def save_service(self, name: str, url: str, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError

    def delete_service(self, name: str) -> None:
        raise NotImplementedError

    def list_services(self) -> List[Service]:
        raise NotImplementedError

    def fetch_by_name(self, name: str) -> Service:
        raise NotImplementedError

    def fetch_by_url(self, url: str) -> Service:
        raise NotImplementedError

    def clear_services(self) -> None:
        raise NotImplementedError


class OWSRegistryInterface(object):
    def register_service(self, name: str, url: str, *args, **kwargs) -> Dict:
        """Register an OWS service with given ``name`` and ``url``."""
        raise NotImplementedError

    def unregister_service(self, name: str) -> bool:
        """Unregister an OWS service with given ``name``."""
        raise NotImplementedError

    def get_service_by_name(self, name: str) -> Dict:
        """Lookup OWS service with given ``name``."""
        raise NotImplementedError

    def get_service_by_url(self, url: str) -> Dict:
        """Lookup OWS service with given ``url``."""
        raise NotImplementedError

    def list_services(self) -> List:
        """List all registered OWS services."""
        raise NotImplementedError

    def clear_services(self) -> bool:
        """Remove all registered OWS services."""
        raise NotImplementedError



from twitcher.utils import get_settings

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pyramid.config import Configurator
    from pyramid.request import Request
    from pyramid.response import Response

    from twitcher.interface import OWSSecurityInterface, OWSRegistryInterface
    from twitcher.models.service import ServiceConfig
    from twitcher.typedefs import AnySettingsContainer, JSON


class AdapterInterface(object):
    """
    Common interface allowing functionality overriding using an adapter implementation.
    """
    def __init__(self, container):
        # type: (AnySettingsContainer) -> None
        self.settings = get_settings(container)

    @property
    def name(self):
        return '{}.{}'.format(self.__module__, type(self).__name__)

    def describe_adapter(self):
        # type: () -> JSON
        """
        Returns a JSON serializable dictionary describing the adapter implementation.
        """
        raise NotImplementedError

    def configurator_factory(self, container):
        # type: (AnySettingsContainer) -> Configurator
        """
        Returns the 'configurator' implementation of the adapter.
        """
        raise NotImplementedError

    def owssecurity_factory(self):
        # type: () -> OWSSecurityInterface
        """
        Returns the 'owssecurity' implementation of the adapter.
        """
        raise NotImplementedError

    def owsregistry_factory(self, request):
        # type: (Request) -> OWSRegistryInterface
        """
        Returns the 'owsregistry' implementation of the adapter.
        """
        raise NotImplementedError

    def owsproxy_config(self, container):
        # type: (AnySettingsContainer) -> None
        """
        Returns the 'owsproxy' implementation of the adapter.
        """
        raise NotImplementedError

    def request_hook(self, request, service):
        # type: (Request, ServiceConfig) -> Request
        """
        Apply modifications onto the request before sending it.

        .. versionadded:: 0.7.0

        Request members employed after this hook is called include:
        - :meth:`Request.headers`
        - :meth:`Request.method`
        - :meth:`Request.body`

        This method can modified those members to adapt the request for specific service logic.
        """
        raise NotImplementedError

    def response_hook(self, response, service):
        # type: (Response, ServiceConfig) -> Response
        """
        Apply modifications onto the response from sent request.

        .. versionadded:: 0.7.0

        The received response from the proxied service is normally returned directly.
        This method can modify the response to adapt it for specific service logic.
        """
        raise NotImplementedError

    def send_request(self, request, service):
        # type: (Request, ServiceConfig) -> Response
        """
        Performs the provided request in order to obtain a proxied response.

        .. versionadded:: 0.8.0

        The operation should consider the service definition to resolve where the
        request redirection should be proxied to, and handle any relevant response
        errors, such as an unauthorized access or an unreachable service.
        """
        raise NotImplementedError



"""
Based on unitests in https://github.com/wndhydrnt/python-oauth2/tree/master/oauth2/test
"""

from .common import BaseTest, dummy_request

from twitcher.store import ServiceStore


class ServiceStoreTestCase(BaseTest):
    def setUp(self):
        super(ServiceStoreTestCase, self).setUp()
        self.init_database()

        self.service_store = ServiceStore(
            dummy_request(dbsession=self.session))

    def test_service_store(self):
        self.service_store.save_service(
            name="flamingo",
            url="http://somewhere.over.the/ocean",
            type="wps",
            auth='token',
            verify=True,
            purl="http://purl/wps"
        )
        service = self.service_store.fetch_by_name(name='flamingo')
        assert service.name == 'flamingo'
        service = self.service_store.fetch_by_url(url='http://somewhere.over.the/ocean')
        assert service.url == 'http://somewhere.over.the/ocean'
        services = self.service_store.list_services()
        assert len(services) == 1
        self.service_store.delete_service(name='flamingo')
        self.service_store.clear_services()

    def test_service_store_insert_or_update(self):
        self.service_store.save_service(
            name="albatross",
            url="http://somewhere.over.the/ocean",
            type="wps",
            auth='token',
            verify=True,
            purl="http://purl/wps"
        )
        # update
        self.service_store.save_service(
            name="albatross",
            url="http://somewhere.over.the/ocean",
            type="wps",
            auth='token',
            verify=True,
            purl="http://purl/wps"
        )
        services = self.service_store.list_services()
        assert len(services) == 1
        self.service_store.clear_services()



from sqlalchemy import (
    Column,
    Integer,
    String,
)
from sqlalchemy.ext.hybrid import hybrid_property
from typing import TYPE_CHECKING, Union

from .meta import Base

if TYPE_CHECKING:
    from twitcher.typedefs import TypedDict

    ServiceConfig = TypedDict("ServiceConfig", {
        "url": str,
        "name": str,
        "type": str,
        "purl": str,
        "auth": str,
        "public": bool,
        "verify": bool
    }, total=True)


class Service(Base):
    __tablename__ = 'services'
    id = Column(Integer, primary_key=True)
    url = Column(String(255), nullable=False)
    name = Column(String(40), unique=True, index=True, nullable=False)
    type = Column(String(40))
    purl = Column(String(255))
    _verify = Column(Integer)  # sqlite does not support Boolean
    auth = Column(String(40))

    @hybrid_property
    def verify(self):
        # type: () -> bool
        if self._verify == 1:
            return True
        return False

    @verify.setter
    def verify(self, verify: Union[bool, int]) -> None:
        self._verify = int(verify)

    @property
    def public(self) -> bool:
        """Return true if public access."""
        return self.auth not in ['token', 'cert']

    def json(self):
        # type: () -> ServiceConfig
        return {
            'url': self.url,
            'name': self.name,
            'type': self.type,
            'purl': self.purl,
            'auth': self.auth,
            'public': self.public,
            'verify': self.verify}



"""init

Revision ID: d39766e64a84
Revises: 
Create Date: 2019-05-10 17:46:04.376145

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd39766e64a84'
down_revision = None
branch_labels = None
depends_on = None

def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('access_tokens',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('token', sa.Text(), nullable=True),
    sa.Column('expires_at', sa.Integer(), nullable=True),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_access_tokens'))
    )
    op.create_index('token_index', 'access_tokens', ['token'], unique=True, mysql_length=255)
    op.create_table('services',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('url', sa.Text(), nullable=True),
    sa.Column('name', sa.Text(), nullable=True),
    sa.Column('type', sa.Text(), nullable=True),
    sa.Column('purl', sa.Text(), nullable=True),
    sa.Column('public', sa.Integer(), nullable=True),
    sa.Column('verify', sa.Integer(), nullable=True),
    sa.Column('auth', sa.Text(), nullable=True),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_services'))
    )
    op.create_index('name_index', 'services', ['name'], unique=True, mysql_length=255)
    # ### end Alembic commands ###

def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('name_index', table_name='services')
    op.drop_table('services')
    op.drop_index('token_index', table_name='access_tokens')
    op.drop_table('access_tokens')
    # ### end Alembic commands ###



import json
import mock

from twitcher.adapter.default import DefaultAdapter
from twitcher.owssecurity import OWSSecurityInterface
from twitcher.owsregistry import OWSRegistry
from twitcher.store import ServiceStore

from ..common import dummy_request
from .base import FunctionalTest


class AdapterWithHooks(DefaultAdapter):
    def owssecurity_factory(self):
        class DummyOWSSecurity(OWSSecurityInterface):
            def verify_request(self, request): return True   # noqa: E704
        return DummyOWSSecurity()

    def request_hook(self, request, service):
        request.headers["X-Hook-Test-Service"] = service["name"]
        return request

    def response_hook(self, response, service):
        # must edit body using text content,
        # json property re-generates from it, cannot set value direct on dict returned by it
        data = json.loads(response.text)
        data["Hook-Test-Service"] = service["name"]
        response.body = json.dumps(data).encode("UTF-8")
        return response


class TestAdapterWithHooks(FunctionalTest):
    @property
    def settings(self):
        adapter_name = '{}.{}'.format(AdapterWithHooks.__module__, AdapterWithHooks.__name__)
        settings = super(TestAdapterWithHooks, self).settings.copy()
        settings.update({
            'twitcher.adapter': adapter_name
        })
        return settings

    def setUp(self):
        super(TestAdapterWithHooks, self).setUp()
        self.init_database()
        service_store = ServiceStore(dummy_request(dbsession=self.session))
        self.reg = OWSRegistry(servicestore=service_store)

        self.test_service_name = "test_adapter_svc"
        self.test_service = {
            'url': 'http://localhost/wps',
            'name': self.test_service_name,
            'type': 'wps',
            'auth': 'token',
            'public': False,
            'verify': True,
            'purl': 'http://myservice/wps'}
        resp = self.reg.register_service(**self.test_service)
        assert resp == self.test_service

        self.config.include('twitcher.owsproxy')
        self.app = self.get_test_app()

    def test_request_response_hooks(self):
        test_request_handle = []

        def mocked_request(method, url, data, headers, **_):
            _req = dummy_request(self.session)
            _req.method = method
            _req.url = url
            _req.headers = headers
            _req.body = data
            test_request_handle.append(_req)
            _resp = _req.response
            _resp.content_type = "application/json"
            _resp.status_code = 200
            _resp.body = json.dumps({"response": "ok"}).encode("UTF-8")
            _resp.content = _resp.body
            _resp.ok = True
            return _resp

        with mock.patch("requests.request", side_effect=mocked_request):
            resp = self.app.get(f'/ows/proxy/{self.test_service_name}?service=wps&request=getcapabilities')
            assert resp.status_code == 200
            assert resp.content_type == "application/json"

        # check added header by request hook
        assert test_request_handle
        assert test_request_handle[0].headers.get("X-Hook-Test-Service") == self.test_service_name

        # check added body content by response hook
        assert resp.json == {"response": "ok", "Hook-Test-Service": self.test_service_name}



"""update client

Revision ID: b8aa09688558
Revises: 0a075bcbeb8f
Create Date: 2019-07-23 16:33:44.773166

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b8aa09688558'
down_revision = '0a075bcbeb8f'
branch_labels = None
depends_on = None

def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('client', sa.Column('_redirect_uris', sa.Text(), nullable=True))
    op.add_column('client', sa.Column('name', sa.String(length=40), nullable=True))
    # ### end Alembic commands ###

def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('client', 'name')
    op.drop_column('client', '_redirect_uris')
    # ### end Alembic commands ###



import logging
from sqlalchemy import engine_from_config
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm import configure_mappers, scoped_session
import zope.sqlalchemy

# import or define all models here to ensure they are attached to the
# Base.metadata prior to any initialization routines
from .service import Service   # noqa: F401
from .oauth import Client  # noqa: F401
from .oauth import Token  # noqa: F401

LOGGER = logging.getLogger("TWITCHER")

# run configure_mappers after defining all of the models to ensure
# all relationships can be setup
configure_mappers()


def get_engine(settings, prefix='sqlalchemy.'):
    return engine_from_config(settings, prefix)


def get_session_factory(engine):
    factory = sessionmaker()
    factory.configure(bind=engine)
    return scoped_session(factory)


def get_tm_session(session_factory, transaction_manager):
    """
    Get a ``sqlalchemy.orm.Session`` instance backed by a transaction.

    This function will hook the session to the transaction manager which
    will take care of committing any changes.

    - When using pyramid_tm it will automatically be committed or aborted
      depending on whether an exception is raised.

    - When using scripts you should wrap the session in a manager yourself.
      For example::

          import transaction

          engine = get_engine(settings)
          session_factory = get_session_factory(engine)
          with transaction.manager:
              dbsession = get_tm_session(session_factory, transaction.manager)

    """
    dbsession = session_factory()
    zope.sqlalchemy.register(
        dbsession, transaction_manager=transaction_manager)
    return dbsession


def add_tm_session(request):
    """ Request method that returns a SQLAlchemy session for a request.

    The SQLAlchemy session is managed by a Zope transaction, unless the request has been generated from a
    webtest.TestApp instance for functional testing. In this case:

    - Inspect request.environ dictionary for the SQLAlchemy session referenced by key db.session. Remove the
        session from the request's environment dictionary and return the session.
    - Use the session factory to generate and return a new SQLAlchemy session if there is no entry for db.session
        in the request environment dictionary.

    The webtest.TestApp instance should configure the environment dictionary as follows:
        `testapp = webtest.TestApp(app, extra_environ={'db.session': session, 'tm.active': True})`

    Setting tm.active to True causes the pyramid_tm tween to bypass generating a transaction for the SQLAlchemy
    session on the request.

    This code is taken from:
    https://groups.google.com/forum/#!topic/pylons-discuss/BZCeM_yejEE

    :param request: Pyramid Request instance
    :return: SQLAlchemy session.
    """
    if 'paste.testing' in request.environ and request.environ['paste.testing'] is True:
        if 'db.session' in request.environ:  # and 'db.tm' in request.environ:
            dbsession = request.environ['db.session']
            del request.environ['db.session']
            return dbsession
    session = get_tm_session(request.registry['dbsession_factory'], request.tm)
    return session


def includeme(config):
    """
    Initialize the model for a Pyramid app.

    Activate this setup using ``config.include('twitcher.models')``.

    """
    LOGGER.debug("Preparing database session.")

    settings = config.get_settings()
    settings['tm.manager_hook'] = 'pyramid_tm.explicit_manager'

    # use pyramid_tm to hook the transaction lifecycle to the request
    config.include('pyramid_tm')

    # use pyramid_retry to retry a request when transient exceptions occur
    config.include('pyramid_retry')

    session_factory = get_session_factory(get_engine(settings))
    config.registry['dbsession_factory'] = session_factory

    # make request.dbsession available for use in Pyramid
    config.add_request_method(
        # r.tm is the transaction manager used by pyramid_tm
        # lambda r: get_tm_session(session_factory, r.tm),
        add_tm_session,
        'dbsession',
        reify=True
    )



import os

import requests
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from twitcher.oauth2 import TOKEN_ENDPOINT, CLIENT_APP_ENDPOINT

import logging
LOGGER = logging.getLogger("TWITCHER")

KEYCLOAK_TOKEN_ENDPOINT = '/auth/realms/demo/protocol/openid-connect/token'


class TwitcherService(object):
    """TwitcherService is a twitcher client to talk to the twitcher service API."""
    def __init__(self, url, username=None, password=None, verify=True):
        self.base_url = url
        self.username = username
        self.password = password
        self.verify = verify

    def add_client_app(self, name=None, redirect_uri=None):
        """Add a client application to twitcher with optional name."""
        name = name or ''
        redirect_uri = redirect_uri or ''
        req_url = "{}{}?name={}&redirect_uri={}".format(
            self.base_url,
            CLIENT_APP_ENDPOINT,
            name,
            redirect_uri)
        resp = requests.get(req_url, auth=(self.username, self.password), verify=self.verify)
        if not resp.ok:
            LOGGER.error("Could not add client app.")
        else:
            return resp.json()

    def fetch_token(self, client_id, client_secret, scope=None, keycloak=False):
        """Get an access token with given scope."""
        scope = scope or 'compute'
        os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'
        if keycloak is True:
            endpoint = KEYCLOAK_TOKEN_ENDPOINT
        else:
            endpoint = TOKEN_ENDPOINT
        token_url = "{}{}".format(self.base_url, endpoint)
        client = BackendApplicationClient(client_id=client_id)
        oauth = OAuth2Session(client=client)
        return oauth.fetch_token(token_url,
                                 scope=scope,
                                 client_id=client_id,
                                 client_secret=client_secret,
                                 include_client_id=True,
                                 verify=self.verify)

    def list_services(self):
        """List all registered OWS services."""
        req_url = "{}/services".format(self.base_url)
        resp = requests.get(req_url, auth=(self.username, self.password), verify=self.verify)
        if not resp.ok:
            LOGGER.error("Could not list services")
        else:
            return resp.json()

    def register_service(self, name, url, data=None):
        """Register a service."""
        data = data or {}
        data.update({"name": name, "url": url})
        req_url = "{}/services".format(self.base_url)
        import json
        resp = requests.post(req_url, data=json.dumps(data),
                             auth=(self.username, self.password),
                             headers={},
                             verify=self.verify)
        if not resp.ok:
            LOGGER.error("Could not add service")
        else:
            return resp.json()

    def clear_services(self):
        """Remove all OWS services."""
        req_url = "{}/services".format(self.base_url)
        resp = requests.delete(req_url,
                               auth=(self.username, self.password),
                               headers={},
                               verify=self.verify)
        if not resp.ok:
            LOGGER.error("Could not clear services")
        else:
            return resp.json()

    def get_service(self, name):
        """Get an OWS service with given name."""
        req_url = "{}/services/{}".format(self.base_url, name)
        resp = requests.get(req_url,
                            auth=(self.username, self.password),
                            verify=self.verify)
        if not resp.ok:
            LOGGER.error("Could not get service")
        else:
            return resp.json()

    def unregister_service(self, name):
        """Remove registered service with given name."""
        req_url = "{}/services/{}".format(self.base_url, name)
        resp = requests.delete(req_url,
                               auth=(self.username, self.password),
                               headers={},
                               verify=self.verify)
        if not resp.ok:
            LOGGER.error("Could not remove service")
        else:
            return resp.json()



"""
The owsproxy is a service which acts as a proxy for registered OWS services.

The implementation of owsproxy is based on `papyrus_ogcproxy <https://github.com/elemoine/papyrus_ogcproxy>`_

See also: https://github.com/nive/outpost/blob/master/outpost/proxy.py
"""
import requests

from pyramid.response import Response
from pyramid.settings import asbool
from typing import TYPE_CHECKING

from twitcher.owsexceptions import OWSAccessForbidden, OWSAccessFailed, OWSException, OWSNoApplicableCode
from twitcher.utils import get_settings, get_twitcher_url, is_valid_url, replace_caps_url

import logging
LOGGER = logging.getLogger('TWITCHER')

if TYPE_CHECKING:
    from typing import Iterator

    from pyramid.config import Configurator
    from pyramid.request import Request
    from requests.models import Response as RequestsResponse

    from twitcher.adapter.base import AdapterInterface
    from twitcher.models.service import ServiceConfig
    from twitcher.typedefs import AnySettingsContainer


allowed_content_types = (
    "application/xml",                       # XML
    "text/xml",
    "text/xml;charset=ISO-8859-1"
    "application/vnd.ogc.se_xml",            # OGC Service Exception
    "application/vnd.ogc.se+xml",            # OGC Service Exception
    # "application/vnd.ogc.success+xml",      # OGC Success (SLD Put)
    "application/vnd.ogc.wms_xml",           # WMS Capabilities
    # "application/vnd.ogc.gml",              # GML
    # "application/vnd.ogc.sld+xml",          # SLD
    "application/vnd.google-earth.kml+xml",  # KML
    "application/vnd.google-earth.kmz",
    "image/png",                             # PNG
    "image/png;mode=32bit",
    "image/gif",                             # GIF
    "image/jpeg",                            # JPEG
    "application/json",                      # JSON
    "application/json;charset=ISO-8859-1",
)

# TODO: configure allowed hosts
allowed_hosts = (
    # list allowed hosts here (no port limiting)
    # "localhost",
)


# requests.models.Response defaults its chunk size to 128 bytes, which is very slow
class BufferedResponse(object):
    def __init__(self, resp):
        # type: (RequestsResponse) -> None
        self.resp = resp

    def __iter__(self):
        # type: () -> Iterator[bytes]
        return self.resp.iter_content(64 * 1024)


def send_request(request, service):
    # type: (Request, ServiceConfig) -> Response
    """
    Send the request to the proxied service and handle its response.
    """

    extra_path = request.matchdict.get('extra_path')
    request_params = request.query_string

    # TODO: fix way to build url
    url = service['url']
    if extra_path:
        url += '/' + extra_path
    if request_params:
        url += '?' + request_params
    LOGGER.debug('url = {}'.format(url))

    # forward request to target (without Host Header)
    h = dict(request.headers)
    h.pop("Host", h)
    h['Accept-Encoding'] = None
    #
    service_type = service.get('type', 'wps')
    service_verify = service.get('verify', True)
    if service_type and (service_type.lower() != 'wps'):
        try:
            resp_iter = requests.request(method=request.method.upper(), url=url, data=request.body, headers=h,
                                         stream=True, verify=service_verify)
        except Exception as e:
            return OWSAccessFailed("Request failed: {}".format(e))

        # Headers meaningful only for a single transport-level connection
        hop_by_hop = ['connection', 'keep-alive', 'public', 'proxy-authenticate', 'transfer-encoding', 'upgrade']
        return Response(app_iter=BufferedResponse(resp_iter),
                        headers={k: v for k, v in list(resp_iter.headers.items()) if k.lower() not in hop_by_hop},
                        status_code=resp_iter.status_code, request=request)
    else:
        try:
            resp = requests.request(method=request.method.upper(), url=url, data=request.body, headers=h,
                                    verify=service_verify)
        except Exception as e:
            return OWSAccessFailed("Request failed: {}".format(e))

        if resp.ok is False:
            if 'ExceptionReport' in resp.text:
                pass
            else:
                return OWSAccessFailed("Response is not ok: {}".format(resp.reason))

        # check for allowed content types
        ct = None
        # LOGGER.debug("headers=", resp.headers)
        if "Content-Type" in resp.headers:
            ct = resp.headers["Content-Type"]
            if not ct.split(";")[0] in allowed_content_types:
                msg = "Content type is not allowed: {}.".format(ct)
                LOGGER.error(msg)
                return OWSAccessForbidden(msg)
        else:
            # return OWSAccessFailed("Could not get content type from response.")
            LOGGER.warning("Could not get content type from response")

        try:
            if ct in ['text/xml', 'application/xml', 'text/xml;charset=ISO-8859-1']:
                # replace urls in xml content
                # ... if public URL is not configured use proxy url.
                if is_valid_url(service.get('purl')):
                    public_url = service['purl']
                else:
                    public_url = request.route_url('owsproxy', service_name=service['name'])
                # TODO: where do i need to replace urls?
                content = replace_caps_url(resp.content, public_url, service['url'])
            else:
                # raw content
                content = resp.content
        except Exception:
            return OWSAccessFailed("Could not decode content.")

        headers = {}
        if ct:
            headers["Content-Type"] = ct
        return Response(content, status=resp.status_code, headers=headers, request=request)


def owsproxy_base_path(container):
    # type: (AnySettingsContainer) -> str
    settings = get_settings(container)
    return settings.get('twitcher.ows_proxy_protected_path', '/ows').rstrip('/').strip()


def owsproxy_base_url(container):
    # type: (AnySettingsContainer) -> str
    twitcher_url = get_twitcher_url(container)
    owsproxy_path = owsproxy_base_path(container)
    return twitcher_url + owsproxy_path


def owsproxy_view(request):
    # type: (Request) -> Response
    service_name = request.matchdict.get('service_name')
    try:
        service = request.owsregistry.get_service_by_name(service_name)
        if not service:
            LOGGER.debug("No error raised but service was not found: %s", service_name)
            raise OWSAccessFailed("Could not find service: {}".format(service_name))
    except Exception as exc:
        LOGGER.debug("Error occurred while trying to retrieve service: %s", service_name, exc_info=exc)
        return OWSAccessFailed("Could not find service: {}".format(service_name))
    try:
        if not request.is_verified:
            raise OWSAccessForbidden("Access to service is forbidden.")
        # since request can be modified by hooks, keep reference to original adapter
        # in order to ensure both request/response operations are handled by the same logic
        adapter = request.adapter
        request = adapter.request_hook(request, service)
        response = adapter.send_request(request, service)
        response = adapter.response_hook(response, service)
        return response
    except OWSException as exc:
        LOGGER.warning("Security check failed but was not handled as expected by 'is_verified' method.", exc_info=exc)
        raise
    except Exception as exc:
        LOGGER.exception("Security check failed due to unhandled error.", exc_info=exc)
        raise OWSNoApplicableCode("Unhandled error: {!s}".format(exc))


def owsverify_view(request):
    # type: (Request) -> Response
    """
    Verifies if request access is allowed, but without performing the proxied request and response handling.
    """
    message, status, access = "forbidden", 403, False
    try:
        service_name = request.matchdict.get('service_name')
        service = request.owsregistry.get_service_by_name(service_name)
        if service and request.is_verified:
            message, status, access = "allowed", 200, True
    except Exception as exc:
        LOGGER.exception("Security check failed due to unhandled error.", exc_info=exc)
        pass
    return Response(
        json={"description": "Access to service is {!s}.".format(message), "access": access},
        status=status,
        request=request,
    )


def owsproxy_defaultconfig(config):
    # type: (Configurator) -> None
    settings = get_settings(config)
    if asbool(settings.get('twitcher.ows_proxy', True)):
        protected_path = owsproxy_base_path(settings)

        config.include('twitcher.oauth2')
        config.include('twitcher.owsregistry')
        config.include('twitcher.owssecurity')
        config.add_route('owsproxy', protected_path + '/proxy/{service_name}')
        config.add_route('owsproxy_extra', protected_path + '/proxy/{service_name}/{extra_path:.*}')
        config.add_route('owsverify', protected_path + '/verify/{service_name}')
        config.add_route('owsverify_extra', protected_path + '/verify/{service_name}/{extra_path:.*}')
        config.add_view(owsproxy_view, route_name='owsproxy')
        config.add_view(owsproxy_view, route_name='owsproxy_extra')
        config.add_view(owsverify_view, route_name='owsverify')
        config.add_view(owsverify_view, route_name='owsverify_extra')


def includeme(config):
    # type: (Configurator) -> None
    from twitcher.adapter import get_adapter_factory

    def get_adapter(request):
        # type: (Request) -> AdapterInterface
        adapter = get_adapter_factory(request)
        return adapter

    get_adapter_factory(config).owsproxy_config(config)
    config.add_request_method(get_adapter, reify=False, property=True, name="adapter")



from twitcher.adapter import get_adapter_factory
from twitcher.utils import get_settings, get_twitcher_url, is_json_serializable
from twitcher.oauth2 import CLIENT_APP_ENDPOINT, TOKEN_ENDPOINT
from twitcher.owsproxy import owsproxy_base_url
from twitcher import __version__
from pyramid.settings import asbool
from pyramid.view import view_config

VERSIONS_PATH = '/versions'
INFORMATION_PATH = '/info'


@view_config(route_name='frontpage', renderer='json')
def frontpage(request):
    url = get_twitcher_url(request)
    body = {
        'message': 'Twitcher Frontpage',
        'information_uri': url + INFORMATION_PATH,
        'versions_uri': url + VERSIONS_PATH,
        'services_uri': "{}/services".format(url),
        'client_uri': "{}{}".format(url, CLIENT_APP_ENDPOINT),
        'token_uri': "{}{}".format(url, TOKEN_ENDPOINT),
        'openapi_uri': "{}/__api__".format(url),
    }
    settings = get_settings(request)
    if asbool(settings.get('twitcher.ows_proxy', True)):
        body["owsproxy_uri"] = owsproxy_base_url(request) + "/proxy"
    else:
        body["owsproxy_uri"] = None
    return body


@view_config(route_name='information', renderer='json')
def information(request):
    """List API information."""
    return dict(
        (i.replace('__', ''), getattr(__version__, i))
        for i in dir(__version__)
        if i not in ['__file__', '__name__', '__cached__'] and is_json_serializable(getattr(__version__, i))
    )


@view_config(route_name='versions', renderer='json')
def versions(request):
    """List version details of components used by the API."""
    adapter_version = get_adapter_factory(request).describe_adapter()
    adapter_version['type'] = 'adapter'
    twitcher_version = {
        'name': 'Twitcher',
        'version': __version__.__version__,
        'type': 'application',
    }
    return [twitcher_version, adapter_version]


def includeme(config):
    config.add_route('frontpage', '/')
    config.add_route('information', INFORMATION_PATH)
    config.add_route('versions', VERSIONS_PATH)
